@@ -4,6 +4,7 @@ import queue
 import re
 import socket
 import threading
+import time
 from typing import Generator, Optional
 
 import openvpn_auth_azure_ad.util.errors as errors
@@ -52,7 +53,11 @@ class OpenVPNManagementInterface(object):
         self._listener_thread = None
         self._writer_thread = None
 
-        self._recv_queue = queue.Queue()
+        self._recv_queue = {
+            "COMMAND": queue.Queue(),
+            "CLIENT": queue.Queue(),
+            "INFO": queue.Queue(),
+        }
         self._send_queue = queue.Queue()
 
     @property
@@ -70,54 +75,68 @@ class OpenVPNManagementInterface(object):
         else:
             return str(self._mgmt_socket)
 
-    def connect(self) -> Optional[bool]:
+    def connect(self, retry: bool = False) -> Optional[bool]:
         """Connect to management interface socket.
         """
-        try:
-            if self.type == SocketType.IP:
-                self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
-                self._socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
-                self._socket.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
-                self._socket.connect(
-                    ("{}".format(self._mgmt_host), int(self._mgmt_port))
-                )
-            else:
-                self._socket = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
-                self._socket.connect(self._mgmt_socket)
+        while True:
+            try:
+                if self.type == SocketType.IP:
+                    self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
+                    self._socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
+                    self._socket.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
+                    self._socket.connect(
+                        ("{}".format(self._mgmt_host), int(self._mgmt_port))
+                    )
+                else:
+                    self._socket = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
+                    self._socket.connect(self._mgmt_socket)
 
-            resp = self._socket.recv(1024).decode("utf-8")
-
-            if resp.startswith("ENTER PASSWORD"):
-                self._socket.send((self._mgmt_password + "\n").encode("utf-8"))
                 resp = self._socket.recv(1024).decode("utf-8")
-                if not resp.startswith("SUCCESS: password is correct"):
-                    logger.critical("Wrong management interface password.")
 
-                assert resp.startswith(
-                    "SUCCESS: password is correct"
-                ), "Wrong management interface password."
-            else:
-                assert resp.startswith(
-                    ">INFO"
-                ), "Did not get expected response from interface when opening socket."
+                self._socket_file = self._socket.makefile("r")
+                self._listener_thread = threading.Thread(
+                    target=self._socket_listener_thread,
+                    daemon=True,
+                    name="mgmt-listener",
+                )
+                self._writer_thread = threading.Thread(
+                    target=self._socket_writer_thread, daemon=True, name="mgmt-writer"
+                )
 
-            self._socket_file = self._socket.makefile("r")
-            self._listener_thread = threading.Thread(
-                target=self._socket_listener_thread, daemon=True, name="mgmt-listener"
-            )
-            self._writer_thread = threading.Thread(
-                target=self._socket_writer_thread, daemon=True, name="mgmt-writer"
-            )
+                self._listener_thread.start()
+                self._writer_thread.start()
 
-            self._listener_thread.start()
-            self._writer_thread.start()
+                if resp.startswith("ENTER PASSWORD"):
+                    resp = self.send_command(self._mgmt_password)
+                    if not resp.startswith("SUCCESS: password is correct"):
+                        logger.critical("Wrong management interface password.")
 
-            self._get_version()
-            logger.info("Connection to OpenVPN management interfaced established.")
+                    assert resp.startswith(
+                        "SUCCESS: password is correct"
+                    ), "Wrong management interface password."
 
-            return True
-        except (socket.timeout, socket.error) as e:
-            raise errors.ConnectError(str(e)) from None
+                    resp = self._socket_recv("INFO")
+                    assert resp.startswith(
+                        ">INFO"
+                    ), "Did not get expected response from interface when opening socket."
+                else:
+                    assert resp.startswith(
+                        ">INFO"
+                    ), "Did not get expected response from interface when opening socket."
+
+                self._get_version()
+                logger.info("Connection to OpenVPN management interfaced established.")
+
+                return True
+            except AssertionError as e:
+                if not retry:
+                    raise e
+            except (socket.timeout, socket.error) as e:
+                if retry:
+                    logger.error(str(e))
+                    time.sleep(1)
+                else:
+                    raise errors.ConnectError(str(e)) from None
 
     def disconnect(self, _quit=True) -> None:
         """Disconnect from management interface socket.
@@ -161,12 +180,21 @@ class OpenVPNManagementInterface(object):
                 break
 
             line = self._socket_file.readline().strip()
+
             if not line:
-                self._recv_queue.put(None)
+                self._recv_queue["CLIENT"].put(None)
                 break
 
-            if line.startswith("SUCCESS:") or line.startswith("ERROR:"):
-                self._recv_queue.put(line + "\n")
+            if line.startswith(">INFO:"):
+                self._recv_queue["INFO"].put(line + "\n")
+                continue
+
+            if (
+                line.startswith("SUCCESS:")
+                or line.startswith("ERROR:")
+                or line.startswith("ENTER PASSWORD")
+            ):
+                self._recv_queue["COMMAND"].put(line + "\n")
                 continue
 
             if len(recv_lines) == 0:
@@ -176,7 +204,8 @@ class OpenVPNManagementInterface(object):
 
             if LAST_LINE_REGEX.match(line):
                 self._socket_io_lock.release()
-                self._recv_queue.put("\n".join(recv_lines))
+                queue_name = "CLIENT" if line.startswith(">CLIENT") else "COMMAND"
+                self._recv_queue[queue_name].put("\n".join(recv_lines))
                 recv_lines = []
 
     def _socket_writer_thread(self):
@@ -200,7 +229,7 @@ class OpenVPNManagementInterface(object):
             )
         self._send_queue.put(data)
 
-    def _socket_recv(self) -> str:
+    def _socket_recv(self, queue_name: str) -> str:
         """Receive bytes from socket and convert to string.
         """
         if self._socket is None:
@@ -208,7 +237,7 @@ class OpenVPNManagementInterface(object):
                 "You must be connected to the management interface to issue commands."
             )
 
-        return self._recv_queue.get()
+        return self._recv_queue[queue_name].get()
 
     def send_command(self, cmd) -> Optional[str]:
         """Send command to management interface and fetch response.
@@ -221,7 +250,7 @@ class OpenVPNManagementInterface(object):
         self._socket_send(cmd + "\n")
         if cmd.startswith("kill") or cmd.startswith("client-kill"):
             return
-        resp = self._socket_recv()
+        resp = self._socket_recv("COMMAND")
         logger.debug("Cmd response: %r", resp)
         return resp
 
@@ -245,7 +274,7 @@ class OpenVPNManagementInterface(object):
             )
         logger.debug("Waiting for incoming data")
 
-        return self._socket_recv()
+        return self._socket_recv("CLIENT")
 
     @staticmethod
     def has_prefix(line) -> bool:
